@@ -18,7 +18,7 @@ class LogCatcher(threading.Thread):
         self._lines = []
         self._remainder = b""
         threading.Thread.__init__(self)
-        self.setDaemon(True)  # do not let this thread hold up Python shutdown
+        self.daemon = True  # do not let this thread hold up Python shutdown
         self._should_stop = False
         self.start()
 
@@ -76,6 +76,13 @@ class LogCatcher(threading.Thread):
             elif self._lines:
                 self._lines = limit_lines_local(self._lines)
 
+        # Close the file when we're done
+        # See #61 and #69
+        try:
+            self._file.close()
+        except Exception:
+            pass
+
 
 def get_output_video_line(lines):
     """Get the line that defines the video stream that ffmpeg outputs,
@@ -174,6 +181,14 @@ def parse_ffmpeg_header(text):
             )
         )
 
+    # get the rotate metadata
+    reo_rotate = re.compile("rotate\s+:\s([0-9]+)")
+    match = reo_rotate.search(text)
+    rotate = 0
+    if match is not None:
+        rotate = match.groups()[0]
+    meta["rotate"] = int(rotate)
+
     # get duration (in seconds)
     line = [l for l in lines if "Duration: " in l][0]
     match = re.search(" [0-9][0-9]:[0-9][0-9]:[0-9][0-9].[0-9][0-9]", line)
